/*
 * Meridian Prime
 * Copyright (C) 2002 Jason Cline
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <windows.h>
#include <stdio.h>
#include <mmsystem.h>
#include "global.h"
#include "equalizer.h"
#include "dsp.h"

#define BUFFERS 2
#define UNUSED  0
#define USED    1

HWAVEOUT hWaveOut=NULL;

typedef struct tagBUFFER{
  INT8     *pBuffer;
  WAVEHDR  tWaveHeader;
  UINT32   nSize;
  UINT32   nWritePos;
}BUFFER;

BUFFER   tBuffer[BUFFERS];
UINT32   nPlaying=0;
UINT32   nPaused=0;
UINT32   nEqualizer=0;
UINT32   nCurrentWrite=0;
UINT32   nSamplerate;
UINT32   nBPS;

//this handles our callback for seamless streaming
static void CALLBACK waveOutProc(HWAVEOUT htWaveOut, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
WAVEHDR *tTempWaveHeader=NULL;

  if(uMsg==WOM_DONE){
    //we are done with this buffer, flag it so we can use it again
    tTempWaveHeader=(WAVEHDR *)dwParam1;
    tTempWaveHeader->dwUser=UNUSED;
  }
} 

//shutdown the sound engine
void audioShutdown()
{
UINT32       i;

  if(hWaveOut){
    waveOutReset(hWaveOut);
    //free the buffers
    for(i=0; i<BUFFERS; i++){
      waveOutUnprepareHeader(hWaveOut, &tBuffer[i].tWaveHeader, sizeof(WAVEHDR));
      if(tBuffer[i].pBuffer) free(tBuffer[i].pBuffer);
      ZeroMemory(&tBuffer[i], sizeof(BUFFER));
    }
    waveOutClose(hWaveOut);
  }
  //reset our variables
  hWaveOut=NULL;
  nPlaying=0;
  nCurrentWrite=0;
}

//start up the sound engine
UINT32 audioStartup(HWND hWnd, UINT32 uSamplerate, UINT16 uBits)
{
WAVEFORMATEX wfx;
UINT32       i;

  //stop anything if it is playing
  audioShutdown();
  //set out output format
  wfx.nSamplesPerSec  = uSamplerate;
  wfx.wBitsPerSample  = uBits;
  wfx.nChannels       = 2;
  wfx.cbSize          = 0;
  wfx.wFormatTag      = WAVE_FORMAT_PCM;
  wfx.nBlockAlign     = wfx.wBitsPerSample / 8 * wfx.nChannels;
  wfx.nAvgBytesPerSec = wfx.nBlockAlign * wfx.nSamplesPerSec;
  //try to open the wave mapper
  if(waveOutOpen(&hWaveOut, WAVE_MAPPER, &wfx, (DWORD_PTR)waveOutProc, 0, CALLBACK_FUNCTION) != MMSYSERR_NOERROR){
    Message("Audio: Could not open wave mapper.");
    return(0);
  }
  //reset the buffers
  for(i=0; i<BUFFERS; i++){
    ZeroMemory(&tBuffer[i], sizeof(BUFFER));
    tBuffer[i].pBuffer=(INT8 *)malloc(wfx.nAvgBytesPerSec / 2);
    tBuffer[i].nSize=wfx.nAvgBytesPerSec / 2;
    tBuffer[i].tWaveHeader.lpData=tBuffer[i].pBuffer;
  }
  //reset our variables
  nPlaying=0;
  nPaused=0;
  nCurrentWrite=0;
  nSamplerate=uSamplerate;
  nBPS=uBits;
  //reset the DSPs
  resetDSPs();
  //and exit
  return(1);
}

//pause or unpause the buffer back up
void audioPause()
{
  //exit if the sound engine isnt started
  if(hWaveOut==NULL) return;
  //unpause it
  if(nPaused){
    nPaused=0;
    waveOutRestart(hWaveOut);
  }
  //pause it
  else{
    nPaused=1;
    waveOutPause(hWaveOut);
  }
}

//turn the equalizer on or off
void audioEqualizer()
{
  nEqualizer=(nEqualizer+1) & 0x1;
}

//change the volume of the sound
void audioSetVolume(UINT32 nVolume)
{
UINT32 uVol;
UINT16  uB;

  //exit if the sound engine isnt started
  if(hWaveOut==NULL) return;
  //compute the volume
  uB=(0xffff * (UINT16)nVolume) / 100;
  uVol=(uB << 16) | uB;
  //set it
//  waveOutSetVolume(hWaveOut, uVol);
}

//check if buffer is playing
UINT32 audioPlaying()
{
  if(hWaveOut==NULL) return(0);
  //if we are paused then we arent playing
  if(nPaused) return(0); 
  if(nPlaying) return(1);
  return(0);
}

//check to see if there is room in the buffer
UINT32 audioCanWrite(UINT32 nSize)
{
  //exit if the sound engine isnt started
  if(hWaveOut==NULL) return(0);
  //answer no if the current write buffer is not played yet
  if(tBuffer[nCurrentWrite].tWaveHeader.dwUser==USED) return(0);
  //otherwise see if there is room
  if(tBuffer[nCurrentWrite].nSize - tBuffer[nCurrentWrite].nWritePos >= nSize) return(1);
  //default to no
  return(0);
}

//write to the audio buffer
void audioWrite(INT8 *pBuffer, UINT32 nSize)
{
UINT32 i;

  //exit if the sound engine isnt started
  if(hWaveOut==NULL) return;
  //process the dsp plugins here
  processDSPs(nSamplerate, nBPS, pBuffer, nSize);
  //process the equalizer here
  if(nEqualizer) eqProcess(pBuffer, nSize);
  else eqAmplify(pBuffer, nSize);
  //unprepare the header if we need to
  if(tBuffer[nCurrentWrite].nWritePos==0 && tBuffer[nCurrentWrite].tWaveHeader.dwFlags==WHDR_DONE) waveOutUnprepareHeader(hWaveOut, &tBuffer[nCurrentWrite].tWaveHeader, sizeof(WAVEHDR));
  //copy the input buffer into the play buffer
  CopyMemory(tBuffer[nCurrentWrite].tWaveHeader.lpData + tBuffer[nCurrentWrite].nWritePos, pBuffer, nSize);
  //move the write position
  tBuffer[nCurrentWrite].nWritePos+=nSize;
  //fill some buffers if we havent started playing yet
  if(nPlaying==0){
    //if the buffer it full, prepare it, after 2 full buffers start playing
    if((tBuffer[nCurrentWrite].nWritePos >= tBuffer[nCurrentWrite].nSize) || (tBuffer[nCurrentWrite].nWritePos + nSize > tBuffer[nCurrentWrite].nSize)){
      tBuffer[nCurrentWrite].tWaveHeader.dwBufferLength=tBuffer[nCurrentWrite].nWritePos;
      tBuffer[nCurrentWrite].tWaveHeader.dwUser=USED;
      waveOutPrepareHeader(hWaveOut, &tBuffer[nCurrentWrite].tWaveHeader, sizeof(WAVEHDR));
      tBuffer[nCurrentWrite].nWritePos=0;
      nCurrentWrite=(nCurrentWrite + 1) % BUFFERS;
    }
    if(nCurrentWrite==BUFFERS / 2){
      for(i=0; i<BUFFERS / 2; i++)
        waveOutWrite(hWaveOut, &tBuffer[i].tWaveHeader, sizeof(WAVEHDR));
      nPlaying=1;
    }
  }
  //otherwise queue up the buffer
  else{
    //if the buffer it full, play it
    if((tBuffer[nCurrentWrite].nWritePos >= tBuffer[nCurrentWrite].nSize) || (tBuffer[nCurrentWrite].nWritePos + nSize > tBuffer[nCurrentWrite].nSize)){
      tBuffer[nCurrentWrite].tWaveHeader.dwBufferLength=tBuffer[nCurrentWrite].nWritePos;
      tBuffer[nCurrentWrite].tWaveHeader.dwUser=USED;
      waveOutPrepareHeader(hWaveOut, &tBuffer[nCurrentWrite].tWaveHeader, sizeof(WAVEHDR));
      waveOutWrite(hWaveOut, &tBuffer[nCurrentWrite].tWaveHeader, sizeof(WAVEHDR));
      tBuffer[nCurrentWrite].nWritePos=0;
      nCurrentWrite=(nCurrentWrite + 1) % BUFFERS;
    }
  }
}